# encoding: utf-8
"""
List categories and their IDs in a Discourse forum.
"""

import os
from argparse import ArgumentParser
from community_mailbot.discourse import SiteFeed


def main():
    args = parse_args()

    site_feed = SiteFeed(args.url, user=args.user, key=args.key)
    for c_id, name in site_feed.category_names.items():
        print(c_id, name)


def parse_args():
    parser = ArgumentParser(description=__doc__)
    parser.add_argument(
        '--key',
        default=os.getenv('DISCOURSE_KEY', None),
        help='Discourse API key')
    parser.add_argument(
        '--user',
        default=os.getenv('DISCOURSE_USER', None),
        help='Discourse API user')
    parser.add_argument(
        '--url',
        default='http://community.lsst.org',
        help='Base URL of the discourse forum')
    return parser.parse_args()


if __name__ == '__main__':
    main()



# encoding: utf-8
"""
Processing tools for main content.
"""
from urllib.parse import urlunsplit, urlparse
from collections import namedtuple

from bs4 import BeautifulSoup


# Compatible with tuples form urllib.parse
URL = namedtuple('URL', 'scheme netloc path query fragment')
URL.__new__.__defaults__ = ('http', '', '', '', '', '')


def clean_discourse_html(html, base_url):
    base_url_parts = urlparse(base_url)
    soup = BeautifulSoup(html, 'html.parser')

    for link_tag in soup.find_all('a'):
        link_tag['href'] = make_absolute_link(link_tag['href'], base_url_parts)

    for link_tag in soup.find_all('img'):
        link_tag['src'] = make_absolute_link(link_tag['src'], base_url_parts)

    return soup.prettify()


def make_absolute_link(discourse_link, base_url_parts):
    """Make any link from Discourse into an Absolute URL.

    Discourse provides link with a '//' prefix (e.g.
    //community.lsst.org/path) or relative, e.g. (/users/jsick).
    """
    url_parts = urlparse(discourse_link, scheme='http')

    # reconstruct URL
    if len(url_parts.netloc) == 0:
        # relative (local) url; add the scheme and netloc
        return urlunsplit(URL(scheme=url_parts.scheme,
                              netloc=base_url_parts.netloc,
                              path=url_parts.path))
    else:
        # treat as a global URL
        # this roundtrip process ensures the scheme is present
        return urlunsplit(URL(scheme=url_parts.scheme,
                              netloc=url_parts.netloc,
                              path=url_parts.path))



# encoding: utf-8
"""
Tools for reading the JSON API from a Discourse site.
"""

import os
from urllib.parse import urlunsplit, urlparse
import json
from collections import namedtuple

import requests


# Compatible with tuples form urllib.parse
URL = namedtuple('URL', 'scheme netloc path query fragment')
URL.__new__.__defaults__ = ('http', '', '', '', '', '')

# Identification for a topic
TopicID = namedtuple('TopicID', 'iid slug')


class SiteFeed(object):
    """Access to Discourse's site.json

    Parameters
    ----------
    base_url : str
        Root URL of the Discourse forum.
    key : str, optional
        Discourse API key, needed for working with private categories.
    user : str, optional
        Discourse API username, needed for working with private categories.
    """
    def __init__(self, base_url, key=None, user=None):
        super(SiteFeed, self).__init__()
        self._base_url = base_url
        self._key = key
        self._user = user

        self._feed = self._fetch_feed()

    @property
    def url(self):
        """JSON feed URL."""
        parts = urlparse(self._base_url, scheme='http', allow_fragments=True)
        return urlunsplit(URL(scheme=parts.scheme,
                              netloc=parts.netloc,
                              path='site.json'))

    def _fetch_feed(self):
        """Get the category's JSON feed and parse it into a Python dict."""
        params = {}
        if (self._key is not None) and (self._user is not None):
            params['api_key'] = self._key
            params['api_user'] = self._user
        r = requests.get(self.url, params=params)
        print(r.url)
        r.raise_for_status()  # can raise requests.exceptions.HTTPError
        return r.json()

    @property
    def category_names(self):
        """dict mapping category ID (int) to the full category name (str)."""
        return {c['id']: c['name'] for c in self._feed['categories']}

    @property
    def category_paths(self):
        """dict mapping category ID (int) to the category's URL path."""
        paths = {}

        # first pass: get only primary categories
        for c in self._feed['categories']:
            if 'parent_category_id' not in c:
                paths[c['id']] = c['slug']

        # second pass: get subcategories
        for c in self._feed['categories']:
            if 'parent_category_id' in c:
                parent_slug = paths[c['parent_category_id']]
                paths[c['id']] = '/'.join((parent_slug, c['slug']))

        return paths


class CategoryFeed(object):
    """JSON feed from a specific category.

    Parameters
    ----------
    category_path : str
        The category's URL path.
    base_url : str
        Root URL of the Discourse forum.
    key : str, optional
        Discourse API key, needed for working with private categories.
    user : str, optional
        Discourse API username, needed for working with private categories.
    """
    def __init__(self, category_path, base_url, key=None, user=None):
        super().__init__()
        self._category_path = category_path
        self._base_url = base_url
        self._key = key
        self._user = user

        self._feed = self._fetch_feed()

    @property
    def url(self):
        """JSON feed URL."""
        parts = urlparse(self._base_url, scheme='http', allow_fragments=True)
        url = URL(scheme=parts.scheme,
                  netloc=parts.netloc,
                  path='c/{0}/l/latest.json'.format(self._category_path))
        return urlunsplit(url)

    def _fetch_feed(self):
        """Get the category's JSON feed and parse it into a Python dict."""
        params = {}
        if (self._key is not None) and (self._user is not None):
            params['api_key'] = self._key
            params['api_user'] = self._user
        r = requests.get(self.url, params=params)
        print(r.url)
        r.raise_for_status()  # can raise requests.exceptions.HTTPError
        return r.json()

    def new_topics(self, cache):
        """New topics in the feed that aren't cached.

        Parameters
        ----------
        cache : :class:`TopicCache`
            A :class:`TopicCache` instance.

        Returns
        -------
        topic_list : list
            A list of new ``TopicID`` tuples with attributes ``iid`` (``int``)
            and ``slug`` (``str``) of topics that aren't in the cache.
        """
        return [TopicID(iid=t['id'], slug=t['slug'])
                for t in self._feed['topic_list']['topics']
                if t['slug'] not in cache]


class TopicFeed(object):
    """JSON feed from a Discourse topic.

    Parameters
    ----------
    topic_slug : str
        Slug of the topic
    topic_id : int
        ID of the topic
    base_url : str
        Root URL of the Discourse forum.
    key : str, optional
        Discourse API key, needed for working with private categories.
    user : str, optional
        Discourse API username, needed for working with private categories.
    """
    def __init__(self, topic_slug, topic_id, base_url,
                 key=None, user=None):
        super().__init__()
        self._id = topic_id
        self._slug = topic_slug
        self._base_url = base_url
        self._key = key
        self._user = user

        self._feed = self._fetch_feed()

    @property
    def url(self):
        """JSON feed URL."""
        parts = urlparse(self._base_url, scheme='http', allow_fragments=True)
        url = urlunsplit(URL(scheme=parts.scheme,
                             netloc=parts.netloc,
                             path='t/{0:d}.json'.format(self._id)))
        return url

    def _fetch_feed(self):
        """Get the topic's JSON feed and parse it into a Python dict."""
        # Surprisingly requests is not adding these parameters to my URL!
        params = {}
        if (self._key is not None) and (self._user is not None):
            params['api_key'] = self._key
            params['api_user'] = self._user
        r = requests.get(self.url, params=params)
        print(r.url)
        r.raise_for_status()  # can raise requests.exceptions.HTTPError
        return r.json()

    @property
    def slug(self):
        """Slug (URL shortname) of the topic."""
        return self._slug

    @property
    def html_url(self):
        """Topic HTML URL."""
        parts = urlparse(self._base_url, scheme='http', allow_fragments=True)
        return urlunsplit(URL(scheme=parts.scheme,
                              netloc=parts.netloc,
                              path='t/{0}'.format(self._slug)))

    @property
    def datetime_iso(self):
        """ISO datetime string of first post in topic"""
        return self._feed['post_stream']['posts'][0]['created_at']

    @property
    def category_id(self):
        """The topic's category (``int`` ID)."""
        return self._feed['category_id']

    @property
    def title(self):
        """The topic's title."""
        return self._feed['title']

    @property
    def first_post_content(self):
        """Content of the topic's first post, as HTML (latest revision)."""
        return self._feed['post_stream']['posts'][0]['cooked']

    @property
    def first_post_author_real_name(self):
        """Display name of the topic's original poster."""
        return self._feed['post_stream']['posts'][0]['display_username']


class TopicCache(object):
    """Cache of topics already seen/processed.

    The cache is stored in a JSON-formatted file. The parent object is a
    ``dict``, keyed by the topic's slug. Value is a dict with fields

    - `created_at` (an ISO datetime string)
    - `category` (a string)

    .. note::

        In the future this JSON cache could be converted into an sqlite DB
        for performance.

    Parameters
    ----------
    cache_path : str
        File path to the cache file. The cache will be created if it does
        not alreay exist.
    """
    def __init__(self, cache_path):
        super(TopicCache, self).__init__()
        self._path = cache_path

        try:
            self._cache = self._load_cache()
        except IOError:
            # start a brand new cache
            self._cache = {}

    def _load_cache(self):
        with open(self._path, 'r') as f:
            cache = json.load(f)
        return cache

    def __len__(self):
        return len(self._cache)

    def __getitem__(self, key):
        return self._cache[key]

    def __contains__(self, key):
        if key in self._cache:
            return True
        else:
            return False

    def add(self, topic_slug, category_id=None, created_at=None):
        """Add a topic to the cache.

        Parameters
        ----------
        topic_slug : str
            The url-friendly shortname for the topic. Topics are keyed in the
            cache by slug.
        category_id : int
            The identifier for the category containing the topic.
        created_at : str
            The ISO datetime string for when the topic was created.
        """
        self._cache[topic_slug] = {'category_id': category_id,
                                   'created_at': created_at}

    def save(self):
        """Save the cache onto the filesystem."""
        dirname = os.path.dirname(self._path)
        if (dirname is not '') and (os.path.exists(dirname) is False):
            os.makedirs(dirname)

        with open(self._path, 'w') as f:
            json.dump(self._cache, f, indent=2)



# encoding: utf-8
"""
Forward new topics in selected Discourse categories to select email addresses
using Mandrill.

This script is intended to be run as a repeated task that pulls new topics
and forward the HTML content of the topic's first post to email recipents
through Mandrill. When topics are forwarded, they are also cached.

Specific Discourse categories can be monitoried, and each category can have
topics forwarded to one or more email recipients. The map file (see below)
figures the email recipeints for selected categories.

Map file
--------

The ``map_file`` argument should be set to the path of a JSON file that
category IDs to email recipients. The format is::

    {
        "<id>": [{"email": "<email address>",
                  "name": "<recipient name>",
                  "type": "to"}],
    }

See ``config.json`` in the ``community_mailbot`` project repository for a
working example.

The ``discourse_categories`` script, also in the project, lets you discover
the category IDs relative to their human-readable names.
"""

import os
import json
from argparse import ArgumentParser, RawDescriptionHelpFormatter

import requests
import mandrill

from community_mailbot.discourse import (CategoryFeed, TopicFeed, TopicCache,
                                         SiteFeed)
from community_mailbot.contentpipe import clean_discourse_html


def main():
    args = parse_args()

    with open(args.map_file, 'r') as f:
        mapping = json.load(f)

    forward_new_topics(mapping, args.cache,
                       args.url,  args.key, args.user,
                       args.mandrill, args.cache_only)


def parse_args():
    cache_default = os.getenv(
        'COMMUNITY_MAILBOT_CACHE',
        os.path.expandvars('$HOME/.community_mailbot_cache'))
    default_url = 'http://community.lsst.org'
    default_mandrill_key = os.getenv('MANDRILL_KEY', None)
    default_discourse_key = os.getenv('DISCOURSE_KEY', None)
    default_discourse_user = os.getenv('DISCOURSE_USER', None)

    parser = ArgumentParser(
        prog='forward_discourse',
        description=__doc__,
        formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument(
        'map_file',
        help='Path to JSON file mapping categories to E-mail addresses')
    parser.add_argument(
        '--cache',
        help='Path to the topic cache file (or $COMMUNITY_MAILBOT_CACHE)',
        default=cache_default)
    parser.add_argument(
        '--key',
        default=default_discourse_key,
        help='Discourse API key (or $DISCOURSE_KEY)')
    parser.add_argument(
        '--user',
        default=default_discourse_user,
        help='Discourse API user (or $DISCOURSE_USER)')
    parser.add_argument(
        '--url',
        default=default_url,
        help='Base URL of the discourse forum')
    parser.add_argument(
        '--mandrill',
        default=default_mandrill_key,
        help='MANDRILL API key (or $MANDRILL_KEY)')
    parser.add_argument(
        '--cache-only',
        default=False,
        action='store_true',
        help='Do not send emails, process and cache topics only')
    return parser.parse_args()


def forward_new_topics(mapping, cache_path,
                       base_url, discourse_key, discourse_user,
                       mandrill_key, cache_only):
    """Coordinating function.

    Each category is processed and the cache is persisted at the end.
    """
    cache = TopicCache(cache_path)
    site_feed = SiteFeed(base_url, key=discourse_key, user=discourse_user)
    category_names = site_feed.category_names
    category_paths = site_feed.category_paths

    for category_id, recipients in mapping.items():
        cat_feed = CategoryFeed(category_paths[int(category_id)], base_url,
                                key=discourse_key, user=discourse_user)
        for topic in cat_feed.new_topics(cache):
            forward_topic(topic.slug, topic.iid, category_id,
                          recipients, cache,
                          base_url, discourse_key, discourse_user,
                          mandrill_key, cache_only,
                          category_names[int(category_id)])
            cache.save()


def forward_topic(topic_slug, topic_id, category_id, recipients, cache,
                  base_url, discourse_key, discourse_user, mandrill_key,
                  cache_only, category_name):
    """Forward a topic post to the destination email addresses.

    The topic is cached once it is successfuly posted.
    """
    print(topic_slug)
    try:
        topic = TopicFeed(topic_slug, topic_id, base_url,
                          key=discourse_key, user=discourse_user)
    except requests.exceptions.HTTPError:
        print('Bad Permissions {0} (skipping)'.format(topic_slug))
        return

    if int(topic.category_id) != int(category_id):
        # ensure topics from sub-categories aren't included
        print('skipping {0}, {1} is not {2}'.format(topic_slug,
              topic.category_id,
              category_id))
        return

    # FIXME testing
    print('Sending', topic.title, 'to', [r['email'] for r in recipients])

    if not cache_only:
        # docs:
        # https://mandrillapp.com/api/docs/messages.python.html#method-send
        # https://mandrill.zendesk.com/hc/en-us/articles/205582487-How-do-I-use-merge-tags-to-add-dynamic-content-

        # Build messsage content
        template_content = []
        global_merge = [
            {'name': 'BODY_CONTENT',
             'content': topic.first_post_content},
            {'name': 'TOPIC_URL',
             'content': topic.html_url},
            {'name': 'TOPIC_AUTHOR',
             'content': topic.first_post_author_real_name}
        ]
        subject = '[{category}] {title}'.format(category=category_name,
                                                title=topic.title)
        cleaned_html = clean_discourse_html(topic.first_post_content, base_url)
        message = {
            'from_email': 'noreply@community.lsst.org',
            'from_name': topic.first_post_author_real_name,
            'to': recipients,
            'preserve_recipients': False,  # don't expose recipients
            'auto_text': True,
            'track_opens': True,
            'track_clicks': False,
            'signing_domain': None,  # TODO
            'subaccount': 'community_mailbot',
            'subject': subject,
            'html': cleaned_html,
            'global_merge_vars': global_merge,
        }

        # Send message with Mandrill
        mandrill_client = mandrill.Mandrill(mandrill_key)
        try:
            result = mandrill_client.messages.send_template(
                template_name='community-mailbot-template-2',
                template_content=template_content,
                message=message,
                async=False)
        except mandrill.Error as e:
            print(result)
            print('A mandrill error occurred: %s - %s' % (e.__class__, e))
            raise
        print(result)

    # Cache the topic only when we know it was sent
    cache.add(topic_slug,
              category_id=topic.category_id,
              created_at=topic.datetime_iso)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# encoding: utf-8
#
# See COPYRIGHT and LICENSE files at the top of the source tree.
#

from setuptools import setup, find_packages
import os


PACKAGENAME = 'community_mailbot'
DESCRIPTION = 'Friendly mail forwarding bot for community.lsst.org'
LONG_DESCRIPTION = ''
AUTHOR = 'Jonathan Sick'
AUTHOR_EMAIL = 'jsick@lsst.org'
LICENSE = 'MIT'
URL = 'https://github.com/lsst-sqre/community_mailbot'


def read(filename):
    full_filename = os.path.join(
        os.path.abspath(os.path.dirname(__file__)),
        filename)
    return open(full_filename).read()

long_description = read('README.rst')


setup(
    name=PACKAGENAME,
    # Versions should comply with PEP440.
    # (http://www.python.org/dev/peps/pep-0440)
    version='0.1.0',
    description=DESCRIPTION,
    long_description=long_description,
    url=URL,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    license=LICENSE,
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3.5',
    ],
    keywords='astronomy discourse email',
    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),
    install_requires=['requests', 'mandrill', 'beautifulsoup4'],
    tests_require=[],

    # package_data={},

    entry_points={
        'console_scripts': [
            'forward_discourse=community_mailbot.scripts.forward_discourse:main',  # NOQA
            'discourse_categories=community_mailbot.scripts.discourse_categories:main',  # NOQA
        ],
    },
)
